<?php
/**
 * WP QuickDraw is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * WP QuickDraw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with WP QuickDraw. If not, see http://www.gnu.org/licenses/gpl.html.
 * 
 * @package     WP_QuickDraw
 * @author      hifipix
 * @copyright   2018 HifiPix, Inc.
 * @license     http://www.gnu.org/licenses/gpl.html  GNU GPL, Version 3
 * @version     1.0.1
 */
 
namespace Wpqd;

use \WP_Query;

// BatchImporter Static Class 
class BatchImporter {

  const BATCH_PROCESS_LIMIT = 5;
  const MAX_TRIES = 3;
  const BATCH_PROCESS_INTERVAL = 30;
  const TRANSIENT_KEY = 'wpqd-batch-import-progress';
  const CRON_HOOK = 'wpqd_import_image';
  const HEALTHCHECK_CRON_HOOK = 'wpqd_import_healthcheck';
  const QUEUE_OPTION_KEY = 'wpqd_batch_import_list';

  protected $start_time = 0;

  // A reference to an instance of this class.
  private static $instance;

  // Returns an instance of this class.
  public static function get_instance() {
    if (null == self::$instance) {
      self::$instance = new BatchImporter();
    }

    return self::$instance;
  }

  public function __construct() {
    add_action( 'redux/options/' . WPQD_OPTIONS . '/saved', array( $this, 'attemptImportAllMedia' ), 10, 2 );
    add_action( 'wp_ajax_wpqd_reimport_all_ajax', array( $this, 'reimportAllAjax' ) );
    add_action( self::CRON_HOOK, array( $this, 'importImage' ), 10, 2 );
    add_action( self::HEALTHCHECK_CRON_HOOK, array( $this, 'performHealthcheck' ), 10, 2 );
    add_action( 'admin_notices', array( $this, 'addAdminNotice' ) );
    add_action( 'wp_ajax_wpqd_dismiss_admin_import_success', array( $this, 'dismissAdminNoticeAjax' ) );
    add_action( 'wp_ajax_wpqd_check_progress', array( $this, 'checkProgressAjax' ) );
    add_action( 'wp_ajax_wpqd_pause_resume_import', array( $this, 'pauseResumeImportAjax' ) );
    add_action( 'wp_ajax_wpqd_reset_queue', array( $this, 'resetQueueAjax' ) );
    add_filter( 'wp_generate_attachment_metadata', array( $this, 'scheduleImport' ), 100, 2) ;
    add_filter( 'cron_schedules', array( $this, 'healthcheck_cron_schedules' ) );
  }

  public function healthcheck_cron_schedules( $schedules ) {
    if( ! isset( $schedules["5min"] ) ){
        $schedules["5min"] = array(
            // 'interval' => 5*60,
            'interval' => 60, // run faster for testing
            'display' => __('Once every 5 minutes'));
    }
    return $schedules;
  }

  /**
   * Import all images if enabled in plugin settings
   * 
   * @param array $plugin_settings
   * @param array $changed_values
   * @param bool  $force
   * @param bool  $verbose
   * @return  int|bool
   */
  public function attemptImportAllMedia( $plugin_settings, $changed_values, $force = false, $verbose = false ) {
    if ( ! $plugin_settings['enabled'] ) {
      return false;
    }

    // Force imageset regeneration when certain admin settings have changed
    if ( $changed_values && ! empty($changed_values) ) {
      $needs_regenerate = array( 'breakpoints', 'crop_trim' );
      $common_values = array_intersect( $needs_regenerate, array_keys($changed_values) );
      if ( ! empty($common_values) ) {
        $force = true;
      }
    }

    return $this->importAllMedia( $force, $verbose );
  }

  /**
   * Ajax-friendly force reimport of all images; output # of images regenerated
   * 
   * @return  void
   */
  public function reimportAllAjax() {
    $result = $this->reimportAll( true );
    echo json_encode( $result );
    wp_die();
  }

  /**
   * Delete transient when dismissing success message (on "X" icon click)
   *
   * @return  void
   */
  public function dismissAdminNoticeAjax() {
    if ( delete_transient( self::TRANSIENT_KEY ) ) {
        echo true;
    } else {
        echo false;
    }
    wp_die();
  }

  /**
   * Check status of import progress transient data
   * 
   * @return  void
   */
  public function checkProgressAjax() {
    echo json_encode( $this->get_transient_nocache() );
    wp_die();
  }

  /**
   * AJAX-friendly pause or resume the current image generation process
   * 
   * @return  void
   */
  public function pauseResumeImportAjax() {
    echo json_encode( $this->pauseResumeImport() );
    wp_die();
  }

  /**
   * AJAX-friendly force clean image queue
   * 
   * @return  void
   */
  public function resetQueueAjax() {
    echo json_encode( $this->clean_queue( true ) );
    wp_die();
  }

  /**
   * Pause current image generation
   * 
   * @return  bool
   */
  public function pauseImport() {
    wp_clear_scheduled_hook( self::CRON_HOOK );
    wp_clear_scheduled_hook( self::HEALTHCHECK_CRON_HOOK );

    $processed = $this->get_transient_nocache();

    if ( ! is_array($processed) || isset($processed['complete']) ) {
      return false;
    }

    if ( isset($processed['paused']) && $processed['paused'] == true ) {
      return true;
    }

    $processed['paused'] = true;
    return set_transient( self::TRANSIENT_KEY, $processed );
  }

  /**
   * Pause or resume the current image generation process
   * 
   * @return  array
   */
  protected function pauseResumeImport() {
    $processed = $this->get_transient_nocache();

    if ( ! is_array($processed) || isset($processed['complete']) ) {
      return $processed;
    }

    if ( isset($processed['paused']) && $processed['paused'] == true ) {
      $processed['paused'] = false;
      set_transient( self::TRANSIENT_KEY, $processed );

      $this->scheduleNextItem();
      $this->scheduleHealthcheckCron();
    } else {
      $processed['paused'] = true;
      set_transient( self::TRANSIENT_KEY, $processed );

      wp_clear_scheduled_hook( self::CRON_HOOK );
      wp_clear_scheduled_hook( self::HEALTHCHECK_CRON_HOOK );
    }

    return $processed;
  }

  /**
   * Force reimport all images; return if import ran (non-verbose) or # of images regenerated (verbose)
   * 
   * @param bool  $verbose
   * @return  int|bool
   */
  public function reimportAll( $verbose = false ) {
    $options = get_option( WPQD_OPTIONS );
    return $this->attemptImportAllMedia( $options, false, true, $verbose );
  }

  /**
   * Initialize the image generation process
   * 
   * @param   array $images
   * @param   bool  $final_batch
   * @return  void
   */
  public function importImage() {
    $images = $this->get_queue();

    if ( empty($images) ) {
        return false;
    }

    $generator = ImageGenerator::get_instance();
    $uploadDir = wp_upload_dir();
    $baseDir = $uploadDir['basedir'];

    $processed = $this->get_transient_nocache();
    $processed['start_time'] = time();
    set_transient( self::TRANSIENT_KEY, $processed );

    // Get image data
    $post_id = key($images);
    $path = array_shift(array_values($images));

    $imageMetadata = new ImageMetadata($post_id);
    $numberOfTries = $imageMetadata->get_try_count();

    // Too many retries, remove current item from queue
    // and return
    if (!empty($numberOfTries) && is_numeric($numberOfTries) && $numberOfTries >= self::MAX_TRIES) {
      // remove from queue and schedule next item
      unset($images[$post_id]);
      $this->update_queue($images);

      $processed = $this->get_transient_nocache();
      $failed = isset( $processed['failed'] ) ? $processed['failed'] : 0;
      $processed['failed'] = ++$failed;
      set_transient( self::TRANSIENT_KEY, $processed );

      if ( $this->is_queue_empty() ) { // Last image in queue failed
        $processed['complete'] = true;
        $processed['error'] = $this->getProcessedErrorMsg();
        set_transient( self::TRANSIENT_KEY, $processed );
        $this->clean_queue();
      } else {
        $this->scheduleNextItem();
      }

      return false;
    }

    $newPath = ltrim(str_replace($baseDir, '', $path), '/');
    $generator->createSizes(array('file' => $newPath), $post_id);

    if ( $imageMetadata->get_imageset_generated() ) {
      unset($images[$post_id]);
      $this->update_queue($images);

      // Update progress in admin status bar
      $processed = $this->get_transient_nocache();
      $current = isset( $processed['current'] ) ? $processed['current'] : 0;
      $processed['current'] = ++$current;
      set_transient( self::TRANSIENT_KEY, $processed );
    }

    if ( $this->is_queue_empty() ) {
      // Indicate completed status in admin status bar
      $processed['complete'] = true;

      if ( $processed['failed'] > 0 ) {
        $processed['error'] = $this->getProcessedErrorMsg();
      }
      
      set_transient( self::TRANSIENT_KEY, $processed );
      $this->clean_queue();
    } else {
      // Schedule next event
      $this->scheduleNextItem();
    }
  }

  /**
   * Get error message for some images not processed
   * 
   * @return  string
   */
  private function getProcessedErrorMsg() {
    return __( sprintf(
      'Some images were not processed. Please review the <a href="%s">Image Status</a> page for more details.',
      get_site_url( null, '/wp-admin/tools.php?page=wpqd_image_status_page' )
    ), 'wpqd' );
  }

  private function scheduleHealthcheckCron() {
    wp_clear_scheduled_hook( self::HEALTHCHECK_CRON_HOOK );
    wp_schedule_event(time(), '5min', self::HEALTHCHECK_CRON_HOOK);
  }

  public function performHealthcheck() {    
    if ( $this->pastExecutionTime() && ! $this->is_queue_empty() ) {
      // Schedule First Image to process
      $this->scheduleNextItem();
    }

  }

  private function pastExecutionTime() {
    $value = $this->get_transient_nocache();
    $process_start_time = $value['start_time'];

    $time_since_start = time() - $process_start_time;
    $limit = ini_get( 'max_execution_time' );

    // Max Execution Time of 0 = infinity
    if ($limit === 0) {
      $limit = 300;   // 5 minutes
    }

    $past_execution_time = (bool) ( $time_since_start > $limit );

    return $past_execution_time;
  }

  /**
   * Schedule importImage() cron event
   * 
   * @return  void
   */
  public function scheduleNextItem() {
     if ( wp_next_scheduled( self::CRON_HOOK ) ) {
      return;
    }
    wp_schedule_single_event( 
      current_time(time()), 
      self::CRON_HOOK
    );
  }

  /**
   * Schedule import of newly-uploaded image
   * 
   * @param   array   $attachment_metadata
   * @param   int     $attachment_id
   * @return  array
   */
  public function scheduleImport( $attachment_metadata, $attachment_id ) {
    $images = array(
      $attachment_id => get_attached_file( $attachment_id )
    );
    $this->add_queue( $images );

    $can_init = self::initTransient( count($images) );
    if ( ! $can_init ) {
      return $attachment_metadata;
    }

    $this->scheduleNextItem();
    $this->scheduleHealthcheckCron();

    return $attachment_metadata;
  }

  /**
   * Get first non-empty queue from `wp_options`
   * 
   * @param   bool  $value_only
   * @return  array|bool
   */
  public function get_queue( $value_only = true ) {
    global $wpdb;

    $sql = sprintf( 'SELECT `option_name`, `option_value` FROM `%1$soptions`
      WHERE `option_name` LIKE "%2$s_%%"
      AND `option_value` != "%3$s"
      ORDER BY SUBSTRING_INDEX(`option_name`, "_", -1) ASC', 
      $wpdb->prefix,
      self::QUEUE_OPTION_KEY,
      serialize( array() )
    );

    foreach ( $wpdb->get_results( $sql, 'ARRAY_A' ) as $result ) {
      $result['value'] = unserialize( $result['option_value'] );
      if ( empty( $result['value'] ) ) {
        continue;
      }

      if ( $value_only ) {
        return $result['value'];
      }
      return $result;
    }
    return false;
  }

  /**
   * Update existing or create new batch process queue
   * 
   * @param   array $images
   * @return  void
   */
  public function update_queue( $images ) {
    $current_queue = $this->get_queue( false );
    if ( ! $current_queue ) {
      return $this->add_queue( $images );
    }

    update_site_option( $current_queue['option_name'], $images );
  }

  /**
   * Create a new batch process queue
   * 
   * @param   array $images
   * @return  void
   */
  public function add_queue( $images ) {
    global $wpdb;

    $sql = sprintf( 'SELECT MAX(SUBSTRING_INDEX(`option_name`, "_", -1)) 
      FROM `%1$soptions`
      WHERE `option_name` LIKE "%2$s_%%"',
      $wpdb->prefix,
      self::QUEUE_OPTION_KEY
    );

    $index = (int) $wpdb->get_var( $sql );
    $key = sprintf( '%1$s_%2$s', self::QUEUE_OPTION_KEY, ++$index );

    update_site_option( $key, $images );
  }

  /**
   * Check if batch process queue is empty
   * 
   * @return  bool
   */
  public function is_queue_empty() {
    return empty( $this->get_queue() );
  }

  /**
   * Delete empty batch process queues from `wp_options`
   * 
   * @param   bool  $force
   * @return  int|bool
   */
  protected function clean_queue( $force = false ) {
    global $wpdb;

    $sql = sprintf( 'DELETE FROM `%1$soptions`
      WHERE `option_name` LIKE "%2$s_%%"',
      $wpdb->prefix,
      self::QUEUE_OPTION_KEY
    );

    if ( ! $force ) {
      $sql .= sprintf(' AND `option_value` = "%s"', serialize( array() ) );
    } else {
      wp_clear_scheduled_hook( self::CRON_HOOK );
      wp_clear_scheduled_hook( self::HEALTHCHECK_CRON_HOOK );
      delete_transient( self::TRANSIENT_KEY );
    }

    return $wpdb->query( $sql );
  }

  /**
   * Get time limit based on PHP's "max_execution_time" so that we can catch a timeout before it occurs
   * 
   * @param   int $batch_size
   * @param   int $i
   * @return  bool|int
   */
  private function getTimeLimit( $batch_size, $i ) {
    $limit = ini_get( 'max_execution_time' );

    if ( $limit === '0' ) {
      return false;
    }

    $progress = $i / $batch_size;
    $multiplier = ( $progress > 0.5 ) ? $progress : 0.5;
    return (int) round( ( $limit * 0.8 ) * $multiplier );
  }

  /**
   * Display admin notice with image import status
   * 
   * @return  void
   */
  public function addAdminNotice() {
    $processed = $this->get_transient_nocache();
    if ( ! $processed || ! isset($processed['total']) || $processed['total'] == 0 ) {
        return;
    }

    $message = sprintf( 
      'Processed <span class="current">%1$s</span> of <span class="total">%2$s</span> images <span class="failed-wrapper">(<span class="failed">%3$s</span> failed)</span>', 
      isset( $processed['current'] ) ? $processed['current'] : 0,
      $processed['total'],
      isset( $processed['failed'] ) ? $processed['failed'] : 0
    );
    ?>
    <div class="notice notice-warning is-dismissible" id="wpqd-import-success">
        <p><strong><?php _e( 'WP QuickDraw Image Processing', 'wpqd' ); ?></strong></p>
        <p id="wpqd-progress-message"><?php _e( $message, 'wpqd' ); ?><span class="wpqd-spinner-inline"></span></p>
        <p>
            <a href="#" id="wpqd-progress-pauser" class="button button-primary button-large">
              <?php if ( isset( $processed['paused'] ) && $processed['paused'] ): ?>
                <?php _e( 'Resume Import', 'wpqd' ); ?>
              <?php else: ?>
                <?php _e( 'Pause Import', 'wpqd' ); ?>
              <?php endif; ?>
            </a>
        </p>
    </div>
    <script>
      jQuery(document).ready( function($) {
        window.hifipixBatchImporter.checkProgress();
      });
    </script>
    <?php
  }

  /**
   * Initialize transient with data about image import status
   * 
   * @param   int   $image_count
   * @param   bool  $reset
   * @return  bool
   */
  private function initTransient( $image_count, $reset = false ) {
    $value = $this->get_transient_nocache();

    if ( is_array($value) && ! isset($value['complete']) ) {
      if ( $reset ) {
        delete_transient( self::TRANSIENT_KEY );
        return set_transient( self::TRANSIENT_KEY, array(
          'current' => 0,
          'failed' => 0,
          'total' => $image_count
        ) );
      }
      $value['total'] = $value['total'] + $image_count;

      return set_transient( self::TRANSIENT_KEY, $value );
    }

    if ( $image_count < 1 ) {
      return false;
    }

    return set_transient( self::TRANSIENT_KEY, array(
      'current' => 0,
      'failed' => 0,
      'total' => $image_count
    ) );
  }

  /**
   * Get progress indicator transient directly from database (bypasses WP cache)
   * 
   * @return  array
   */
  private function get_transient_nocache() {
    global $wpdb;

    $value = $wpdb->get_var( $wpdb->prepare( 
      "SELECT option_value FROM $wpdb->options WHERE option_name = %s LIMIT 1", 
      '_transient_' . self::TRANSIENT_KEY 
    ) );

    if ( $value ) {
      $value = unserialize( $value );
    }

    return $value;
  }

  /**
   * Import images into WPQD
   * 
   * @param   bool  $force
   * @param   bool  $verbose
   * @return  array|bool
   *   
   * @TODO: store somewhere if and when full import happened
   */
  private function importAllMedia( $force = false, $verbose = false ) {
    $images_result = self::getAllMediaPaths( $force );
    $images = $images_result['images'];
    $can_init = self::initTransient( $images_result['image_count'], true );

    if ( ! $can_init ) {
      return false;
    }

    if ( $force ) {
      $ids = array_keys($images);
      $this->resetImagesetGenerated( $ids );
      $this->resetTryCount( $ids );
    }

    // Add to queue, schedule action with images
    $this->add_queue( $images );

    // Schedule First Image to process
    // and Healthcheck cron
    $this->scheduleNextItem();
    $this->scheduleHealthcheckCron();

    if ( $verbose ) {
        return array(
          'images' => $images_result['image_count']
        );
    }

    return true;
  }

  /**
   * Set "wpqd_generated" = 0 for all images before regeneration
   * 
   * @param   array $ids
   * @return  int|bool
   */
  private function resetImagesetGenerated( $ids ) {
    global $wpdb;

    $query = sprintf(
        'UPDATE `%1$s` SET `meta_value` = 0
        WHERE `meta_key` = "%2$s" AND `post_id` IN (%3$s)',
        _get_meta_table( 'post' ),
        ImageMetadata::KEY_IMAGESET_GENERATED,
        implode(',', $ids)
    );

    return $wpdb->query( $query );
  }

  /**
   * Set "wpqd_try_count" = 0 for all images before regeneration
   * 
   * @param   array $ids
   * @return  int|bool
   */
  private function resetTryCount( $ids ) {
    global $wpdb;

    $query = sprintf(
        'UPDATE `%1$s` SET `meta_value` = 0
        WHERE `meta_key` = "%2$s" AND `post_id` IN (%3$s)',
        _get_meta_table( 'post' ),
        ImageMetadata::KEY_TRY_COUNT,
        implode(',', $ids)
    );

    return $wpdb->query( $query );
  }

  /**
   * Get image paths that require import
   * 
   * @param   bool  $force
   * @return  array
   */
  private static function getAllMediaPaths( $force = false ) {
    $args = array(
      'post_type' => 'attachment',
      'post_mime_type' =>'image',
      'post_status' => 'inherit',
      'post_mime_type' => 'image/jpeg,image/gif,image/jpg,image/png',
      'posts_per_page' => -1,
      'orderby' => 'date',
      'meta_query' => array(
        array(
          'key' => ImageMetadata::KEY_ENABLED,
          'value' => 1,
          'compare' => '='
        )
      )
    );

    if ( ! $force ) {
      // Exclude imagesets that have already been generated
      $args['meta_query'][] = array(
        'key' => ImageMetadata::KEY_IMAGESET_GENERATED,
        'value' => 1,
        'compare' => '!='
      );
      // Exclude imagesets where tries exceeded the maximum
      $args['meta_query'][] = array(
        'key' => ImageMetadata::KEY_TRY_COUNT,
        'value' => self::MAX_TRIES,
        'compare' => '<'
      );
    }

    $query_images = new WP_Query( $args );
    $images = array();
    foreach ( $query_images->posts as $image) {
      $images[$image->ID] = get_attached_file($image->ID);
    }

    return array(
      'image_count' => $query_images->found_posts,
      'images' => $images
    );
  }
}
