<?php
/**
 * WP QuickDraw is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * WP QuickDraw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with WP QuickDraw. If not, see http://www.gnu.org/licenses/gpl.html.
 * 
 * @package     WP_QuickDraw
 * @author      hifipix
 * @copyright   2018 HifiPix, Inc.
 * @license     http://www.gnu.org/licenses/gpl.html  GNU GPL, Version 3
 * @version     1.0.1
 */
 
namespace Wpqd;

//-----------------------------------------
// Disallow Direct Plugin Browsing
//-----------------------------------------
defined( 'ABSPATH' ) or die( 'No script kiddies pretty please!' );

// Image abstraction
require_once(WPQDROOT . 'wpqd-image-editor.php');

// Includes WP Filesystem
require_once(ABSPATH . 'wp-admin/includes/file.php');

// HifipixResize singleton
//
// Orchestrates resizing of images to add HifiPix support
// generateSizes autofires on wordpress upload
class ImageGenerator {

  private $sourceImage;

  /**
   * @var string
   */
  protected $inputFilePath;

  /**
   * @var string
   */
  protected $wpqdUploadsDir;

  /**
   * @var string
   */
  protected $newRootFilename;

  /**
   * @var array
   */
  protected $files;

  const MIN_WIDTH = 256;
  const MIN_HEIGHT = 256;
  const AR_TOLERANCE = 0.01;

  // A reference to an instance of this class.
  private static $instance;

  // Returns an instance of this class.
  public static function get_instance() {
    if (null == self::$instance) {
      self::$instance = new ImageGenerator();
    }
    return self::$instance;
  }

  public function __construct() {
    add_action( 'delete_attachment', array($this, 'deleteSizes'), 10, 1 );
  }

  public function createSizes($attachment_metadata, $attachment_id) {
    return $this->generateSizes($attachment_metadata, $attachment_id, 'create');
  }

  public function updateSizes($attachment_metadata, $attachment_id) {
    return $this->generateSizes($attachment_metadata, $attachment_id, 'update');
  }

  /**
   * Delete WPQD-ified images when the attachment post is deleted
   * 
   * @param   int $attachment_id
   * @return  void
   */
  public function deleteSizes( $attachment_id ) {
    WP_Filesystem(); // Init global $wp_filesystem
    global $wp_filesystem;

    $image_metadata = new ImageMetadata( $attachment_id );
    $root_filename = $image_metadata->get_root_filename();
    $extension = $image_metadata->get_file_extension();
    $pattern = sprintf( '/^%1$s_(\d+_\d{1}_\d+|base)\.%2$s/', $root_filename, $extension );

    $upload_dir = wp_upload_dir();
    $dir = $upload_dir['basedir'] . WPQD_UPLOADS_FOLDER;
    $files = $wp_filesystem->dirlist( $dir );

    foreach ( $files as $filename => $data ) {
      if ( preg_match($pattern, $filename) !== 1 ) {
        continue;
      }
      $wp_filesystem->delete( trailingslashit($dir) . $filename );
    }
  }

  // generateSizes
  //
  private function generateSizes($attachment_metadata, $attachment_id, $context) {

    // error_log(var_export($attachment_metadata, true));
    // error_log(var_export($attachment_id, true));
    // die();

    $uploadDir = wp_upload_dir();

    // TODO: make it support new files with $uploadDir['path'] and old files with $attachment metadata basename
    $this->inputFilePath = trailingslashit($uploadDir['basedir']) . $attachment_metadata['file'];
    $image_metadata = new ImageMetadata($attachment_id);
    $image_metadata->set_try_start_time();
    $image_metadata->set_message();
    
    $tryCount = $image_metadata->get_try_count();
    $image_metadata->set_try_count( ++$tryCount );

    // TODO: Handle update
    // if ($context == 'update') {
    //    delete existing images, then regen from modified original
    // }

    $this->sourceImage = new ImageEditor($this->inputFilePath);
    if ( $message = $this->sourceImage->getErrorMessage() ) {
      $image_metadata->set_imageset_generated( false );
      $image_metadata->set_message( $message );
      return $attachment_matadata;
    }

    $attachmentSize = $this->sourceImage->getSize();

    // Get configured breakpoints from plugin store
    $wpqd_options = get_option(WPQD_OPTIONS);
    $breakpoints = $wpqd_options['breakpoints'];

    // Get paths
    $this->wpqdUploadsDir = trailingslashit( $uploadDir['basedir'] . WPQD_UPLOADS_FOLDER );
    $inputUploadDirectory = dirname($this->inputFilePath);

    // Check if file type and if file exists
    // TODO: Check image type against extension list also
    // so that we can support correctly typed images
    // without the correct extension.
    $acceptedFileTypes = array('gif', 'jpeg', 'jpg', 'png');
    $inputFileInfo = pathinfo($this->inputFilePath);
    $inputFileName = $inputFileInfo['filename'];
    $inputFileExtension = $inputFileInfo['extension'];

    $isAcceptedFiletype = in_array(strtolower($inputFileExtension), $acceptedFileTypes);
    $inputFileExists = file_exists($this->inputFilePath);

    // Store input image size
    $ImageWidth = $attachmentSize['width'];
    $ImageHeight = $attachmentSize['height'];
    $imageIsLargeEnough = $this->imageIsLargeEnough($ImageWidth, $ImageHeight);

    // Disable WPQD for images that are too small or wrong filetype
    if ( ! $inputFileExists || ! $isAcceptedFiletype || ! $imageIsLargeEnough ) {
      $image_metadata->set_enabled( false );

      if ( ! $imageIsLargeEnough ) {
        $image_metadata->set_under_min_limit( true );
      }
      return $attachment_metadata;
    }

    // TODO: Refactor this piece so that event handler gets image and
    // looks at file extension, then routes it here to ppng accordingly
    // HANDLE PPNG
    if ( $inputFileExtension === 'png' && class_exists( '\WpqdPro\ProgressivePNGGenerator' ) ) {
      $ppng_generator = \WpqdPro\ProgressivePNGGenerator::get_instance();
      $ppng_generator->generatePPNG($this->inputFilePath, $attachment_id);
      return $attachment_metadata;
    }

    // Defaults
    $newWidth = $ImageWidth;
    $newHeight = $ImageHeight;

    // Sort Breakpoints largest to smallest
    // so that resizing can happen progressively
    // on the same image
    rsort($breakpoints);

    $aspectRatio = $newWidth / $newHeight;
    $possibleSizes = $this->getPossibleSizes( $newWidth, $aspectRatio );

    // Crop image if we have too few possible resizes
    if ( count($possibleSizes) < 6 ) {
      $moduloWidth = $ImageWidth % 8;
      $moduloHeight = $ImageHeight % 8;

      // Trim
      $newWidth = $ImageWidth - $moduloWidth;
      $newHeight = $ImageHeight - $moduloHeight;

      $image_metadata->set_cropped( true );
      $aspectRatio = $newWidth / $newHeight;  // Get aspect ratio of trimmed image
      $possibleSizes = $this->getPossibleSizes( $newWidth, $aspectRatio, true );
    }

    if ( empty( $possibleSizes ) ) {
      $message = $this->getResizeError( end($possibleSizes) );

      $image_metadata->set_imageset_generated( false );
      $image_metadata->set_message( $message );
      return $attachment_metadata;
    }

    $maxBreakpoint = false;
    $this->files = array();

    // Iterate over breakpoints
    foreach( $breakpoints as $breakpoint_index => $breakpoint_width ) {
      $compareWidth = $breakpoint_width;

      // If it's the last breakpoint & image is still too small, generate images anyway
      if ( ($breakpoint_index + 1 === count($breakpoints)) && ($newWidth < $breakpoint_width) ) {
        $compareWidth = $newWidth;
      }

      // Don't resize if image is smaller than breakpoint width
      if ($newWidth < $compareWidth) {
        continue;
      }

      if ( ! $maxBreakpoint ) {
        $maxBreakpoint = $breakpoint_width;
      }

      // Find appropriate image size for this breakpoint
      foreach ( $possibleSizes as $i => $size ) {
        if ( $size['width'] <= $compareWidth ) {
          continue;
        }

        if ( count($possibleSizes) > 4 ) {
          unset( $possibleSizes[$i] );
        }
      }
      // Reset array indices
      $possibleSizes = array_merge( $possibleSizes );

      // Generate 1/1, 1/2, 1/4, 1/8 sizes
      $sizes = $this->getBreakpointSizes( $possibleSizes, $breakpoint_width );

      if ( $breakpoint_index + 1 === count($breakpoints) ) {
          // Name last image in $sizes $sizes['base']
          $baseImage = end($sizes);
          $baseKey = key($sizes);
          $sizes['base'] = $baseImage;
      }

      $sizes = $this->validateSizes( $sizes, $ImageWidth, $ImageHeight );

      // RESIZE
      $resizedFiles = $this->sourceImage->resize($sizes);
      if ( empty( $resizedFiles ) ) {
        $message = $this->getResizeError( end($sizes) );
        $image_metadata->set_imageset_generated( false );
        $image_metadata->set_message( $message );
        return $attachment_metadata;
      }

      // Move files to /wp-content/uploads/wpqd
      // and rename with metadata scheme
      foreach ($resizedFiles as $breakpoint_name => $f) {
        $filename = $f['file'];
        $originalPath = trailingslashit($inputUploadDirectory) . $filename;

        // TODO: Store root name of file in metadata

        // Construct new filename
        $newPath = $this->getNewFilepath( $filename, $breakpoint_name );

        // TODO: Refactor here to rename() as recursive function
        // which enumerates and checks again until
        // it has a non existant filename

        // If duplicate file, enumerate root filename
        if ( file_exists($newPath) && ! $image_metadata->get_generated_time() ) {
          if (preg_match('/-[1-9]\d*$/', $this->newRootFilename, $matches)) {
            $fileNumber = (int) $matches[0];
            $fileNumber = $fileNumber + 1;
            $dashedFileNumber = '-' . $fileNumber;
            preg_replace('/-[1-9]\d*$/', $dashedFileNumber, $this->newRootFilename);
          } else {
            $dashedFileNumber = '-' . 1;
            $this->newRootFilename .= $dashedFileNumber;
          }

          // Name files
          $filePathInfo = pathinfo( $filename );
          $enumeratedFilename = $this->newRootFilename . "_" . $breakpoint_name . "." . $filePathInfo['extension'];
          $newFilename = $enumeratedFilename;
          $newPath = $this->wpqdUploadsDir . $newFilename;

          // TODO: This is where we check if file exists again and then call self again if so
          // for file enumeration
        }

        // Move file
        // TODO: Decide which files to copy and which to move????

        // If original file, copy, else move
        // Compare with OG file list?

        copy($originalPath, $newPath);
        $this->files[] = $newPath;
      } // file moving loop 
    } // breakpoints loop

    $image_metadata->set_file_extension($inputFileExtension);
    error_log('set_file_extension: ' . $inputFileExtension);

    // Update metadata root with: enumeratedRootFilename
    $image_metadata->set_root_filename($this->newRootFilename);
    error_log('set_root_filename: ' . $this->newRootFilename);

    if ( ! $this->validateFiles( $breakpoints, $maxBreakpoint, $inputFileExtension ) ) {
      // Something went wrong, clear out imageset data + files
      $image_metadata->set_imageset_generated( false );
      $this->deleteSizes( $attachment_id );

      return $attachment_metadata;
    }

    // Upload to CDN
    if ( class_exists( '\WpqdPro\CdnUploader' ) ) {
      $cdn = new \WpqdPro\CdnUploader();
      $cdn->upload( $this->files );
    }

    // Store Aspect Ratio in Metadata
    $image_metadata->set_aspectratio($aspectRatio);
    error_log('set_aspectratio: ' . $aspectRatio);

    // Set largest possible size for image
    $image_metadata->set_max_breakpoint($maxBreakpoint);

    // Update set_imageset_generated true
    $image_metadata->set_imageset_generated(true);
    $image_metadata->set_generated_time();

    $image_metadata->set_enabled(true);
    $image_metadata->set_try_count();

    // Return $upload so it is passed on to the rest
    // of the upload process
    return $attachment_metadata;
  } // generateSizes()

  /**
   * Validate sizes can be resized by WP; if not, copy original instead
   * 
   * @param   array  $sizes
   * @param   int    $width
   * @param   int    $height
   * @return  array
   */
  protected function validateSizes( $sizes, $width, $height ) {
    foreach ( $sizes as $name => $size ) {
      if ( ((int) $size['width']) !== ((int) $width) || 
           ((int) $size['height']) !== ((int) $height) ) {
        continue;
      }

      $this->copyOriginalImage( $name );
      unset( $sizes[$name] );
    }
    return $sizes;
  }

  /**
   * Copy original image to WPQD uploads path
   * 
   * @param   string  $name
   * @return  bool
   */
  protected function copyOriginalImage( $name ) {
    $new_path = $this->getNewFilePath( $this->inputFilePath, $name );
    if ( $copied = copy( $this->inputFilePath, $new_path ) ) {
      $this->files[] = $new_path;
    }
    return $copied;
  }

  /**
   * Build new filepath for image in WPQD uploads directory
   * 
   * @param   string  $filename
   * @param   string  $breakpoint
   * @return  string
   */
  protected function getNewFilepath( $filename, $breakpoint ) {
    $filePathInfo = pathinfo( $filename );
    $this->newRootFilename = preg_replace( '/-[1-9]\d*x[1-9]\d*/', '', $filePathInfo['filename'] );
    $newFilename = $this->newRootFilename  . "_" . $breakpoint . "." . $filePathInfo['extension'];
    return $this->wpqdUploadsDir . $newFilename;
  }

  /**
   * Validate that the files we should have generated, are generated
   * 
   * @param array   $breakpoints
   * @param int     $max
   * @param string  $extension
   * @return  bool
   */
  protected function validateFiles( $breakpoints, $max, $extension ) {
    $upload_dir = wp_upload_dir();
    $wpqd_dir = trailingslashit( $upload_dir['basedir'] . WPQD_UPLOADS_FOLDER );
    $denoms = array( 1, 2, 4, 8 );
    $expected_files = array(
      sprintf( '%1$s%2$s_base.%3$s', $wpqd_dir, $this->newRootFilename, $extension )
    );

    foreach ( $breakpoints as $i => $bp ) {
      if ( $bp > $max ) {
        continue;
      }

      foreach ( $denoms as $denom ) {
        $expected_files[] = sprintf( 
          '%1$s%2$s_%3$s_1_%4$s.%5$s', $wpqd_dir, $this->newRootFilename, $bp, $denom, $extension
        );
      }
    }

    // Check for filename differences between actual & expected
    $diff = array_diff( $expected_files, $this->files );
    if ( ! empty( $diff ) ) {
      return false;
    }

    // Check that expected files exist in filesystem
    foreach ( $expected_files as $file ) {
      if ( ! file_exists( $file ) ) {
        return false;
      }
    }

    return true;
  }

  protected function getResizeError( $size ) {
    if ( $fp = fopen( $this->inputFilePath, 'w' ) ) {
      fclose( $fp );
    } else {
      return __( 'Unable to open source file for writing', 'wpqd' );
    }

    if ( ! $size ) {
      return __( 'Unable to calculate sizes for this image', 'wpqd' );
    }

    if ( $message = $this->imageEditor->testResize( $size ) ) {
      return $message;
    }

    return __( 'Unknown error resizing this image', 'wpqd' );
  }

  /**
   * Get possible sizes for image within .01 threshold
   * 
   * @param   int   $width
   * @param   float $ar
   * @param   bool  $cropped
   * @return  array
   */
  private function getPossibleSizes( $width, $ar, $cropped = false ) {
    $possibleSizes = array();
    $arParts = explode( '.', $ar );
    $decimalPlaces = ( count($arParts) > 1 ) ? strlen( $arParts['1'] ) : 1;
    $factor = (50 / ( $decimalPlaces * 50 )) * 10;
    $decrementBy = ( $factor >= 1 ) ? floor($factor) : ceil($factor);
    $minWidth = ( $width >= 5 * self::MIN_WIDTH ) ? self::MIN_WIDTH : $width / 5;

    for ( $x = $width; $x >= $minWidth; $x -= $decrementBy ) {
      $currentHeight = $x / $ar;
      if ( abs($currentHeight - round($currentHeight)) < self::AR_TOLERANCE ) {
        $currentHeight = (int) round($currentHeight);
      }

      if ( is_int( $currentHeight ) ) {
        $possibleSizes[$x] = array(
          'width' => $x,
          'height' => $currentHeight
        );
      }
    }

    $smallest = end( $possibleSizes );
    $minPossibleWidth = ( $smallest['width'] > $smallest['height'] ) ? $smallest['height'] : $smallest['width'];

    if ( ! $cropped && $minPossibleWidth > ( $minWidth * 1.5 ) ) {
      return array();
    }

    ksort($possibleSizes);
    return array_merge($possibleSizes);
  }

  /**
   * Returns 4 image sizes (1/1, 1/2, 1/4, & 1/8) from set of possibilities
   * 
   * @param   array $possibleSizes
   * @return  array
   */
  private function getBreakpointSizes( $possibleSizes, $bp ) {
    $sizes = array();

    if ( empty($possibleSizes) ) {
      return $sizes;
    }

    $count = count($possibleSizes) - 1; // Subtract 1 to account for 0-indexed array

    $divisor = ( $count > 4 ) ? 4 : $count;
    $indices = array(
      '1' => $count,
      '2' => round( $count - ($count / $divisor ) ),
      '4' => round( $count / $divisor ),
      '8' => 0
    );

    foreach ( $indices as $denom => $index ) {
      $imageSize = $possibleSizes[$index];
      $imageSize['crop'] = true;
      $sizes[$bp . '_1_' . $denom] = $imageSize;
    }

    return $sizes;
  }

  /**
   * Update postmeta values for existing attachment
   *
   * @param     int $id
   * @return    void
   */
  // public function updateMetadata( $id ) {
  //     $metadata = wp_get_attachment_metadata( $id );
  //     $this->saveMetadata( $metadata, $id );
  // }

  private function imageIsLargeEnough($width, $height) {
    return (bool) ($width >= self::MIN_WIDTH && $height >= self::MIN_HEIGHT);
  }

  /**
   * Save root filename, file extension, aspect ratio as postmeta values
   *
   * @param     array   $metadata
   * @param     int     $id
   * @return    array
   */
  public function saveMetadata( $metadata, $id ) {
      $image = wp_prepare_attachment_for_js( $id );
      $image_metadata = new ImageMetadata( $id );

      // if ( isset($image['name']) && ! empty($image['name']) ) {
      //     $image_metadata->set_root_filename( $image['name'] );
      // }

      // if ( isset($image['subtype']) && ! empty($image['subtype']) ) {
      //     $image_metadata->set_file_extension( $image['subtype'] );
      // }

      // if ( isset($metadata['width']) && isset($metadata['height']) ) {
      //     $image_metadata->set_aspectratio( array($metadata['width'], $metadata['height']) );
      // }

      return $metadata;
  }
} // Class ImageGenerator
