<?php
/**
 * WP QuickDraw is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * WP QuickDraw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with WP QuickDraw. If not, see http://www.gnu.org/licenses/gpl.html.
 * 
 * @package     WP_QuickDraw
 * @author      hifipix
 * @copyright   2018 HifiPix, Inc.
 * @license     http://www.gnu.org/licenses/gpl.html  GNU GPL, Version 3
 * @version     1.0.1
 */
 
namespace Wpqd;

//-----------------------------------------
// Disallow Direct Plugin Browsing
//-----------------------------------------
defined( 'ABSPATH' ) or die( 'No script kiddies please!' );

class ImageMetadata {

    /**
     * @var string
     */
    const KEY_ENABLED = 'wpqd_enabled';

    /**
     * @var string
     */
    const KEY_IMAGESET_GENERATED = 'wpqd_generated';

    /**
     * @var string
     */
    const KEY_ROOT_FILENAME = 'wpqd_root_filename';

    /**
     * @var string
     */
    const KEY_FILE_EXT = 'wpqd_file_extension';

    /**
     * @var string
     */
    const KEY_ASPECT_RATIO = 'wpqd_aspectratio';

    /**
     * @var string
     */
    const KEY_MAX_BREAKPOINT = 'wpqd_max_breakpoint';

    /**
     * @var string
     */
    const KEY_UNDER_MIN_LIMIT = 'wpqd_under_min_limit';

    /**
     * @var string
     */
    const KEY_TRY_COUNT = 'wpqd_try_count';

     /**
     * @var string
     */
    const KEY_TRY_START_TIME = 'wpqd_try_start_time';

     /**
     * @var string
     */
    const KEY_GENERATED_TIME = 'wpqd_generated_time';

     /**
     * @var string
     */
    const KEY_PPNG_LEADCOLOR = 'wpqd_ppng_leadcolor';

    /**
     * @var string
     */
    const KEY_IMG_CROPPED = 'wpqd_img_cropped';

    /**
     * @var string
     */
    const KEY_MESSAGE = 'wpqd_messages';

    /**
     * @var array
     */
    protected $key_map;

    /**
     * @var int
     */
    protected $attachment_id;

    /**
     * PHP class constructor
     *
     * @param   int $attachment_id
     */
    public function __construct( $attachment_id ) {
        $this->attachment_id = $attachment_id;
        $this->key_map = array(
            self::KEY_ENABLED => 'set_enabled',
            self::KEY_IMAGESET_GENERATED => 'set_imageset_generated',
            self::KEY_ROOT_FILENAME => 'set_root_filename',
            self::KEY_FILE_EXT => 'set_file_extension',
            self::KEY_ASPECT_RATIO => 'set_aspectratio',
            self::KEY_MAX_BREAKPOINT => 'set_max_breakpoint',
            self::KEY_UNDER_MIN_LIMIT => 'set_under_min_limit',
            self::KEY_TRY_COUNT => 'set_try_count',
            self::KEY_TRY_START_TIME => 'set_try_start_time',
            self::KEY_GENERATED_TIME => 'set_generated_time',
            self::KEY_PPNG_LEADCOLOR => 'set_ppng_leadcolor',
            self::KEY_IMG_CROPPED => 'set_cropped'
        );
    }

    /**
     * Update multiple meta values
     *
     * @param   array   $values
     */
    public function update( $values ) {
        foreach ( $values as $key => $value ) {
            if ( in_array( $key, array_keys($this->key_map) ) ) {
                call_user_func_array( array($this, $this->key_map[$key]), array($value) );
                continue;
            }
            update_post_meta( $this->attachment_id, $key, $value );
        }
    }

    /**
     * Set 'root_filename' meta value
     *
     * @param   string  $filename
     * @return  int|bool
     */
    public function set_root_filename( $filename ) {
        if ( empty($filename) ) {
            return false;
        }
        return update_post_meta( $this->attachment_id, self::KEY_ROOT_FILENAME, $filename );
    }

    /**
     * Get 'root_filename' meta value
     *
     * @return  string
     */
    public function get_root_filename() {
        return get_post_meta( $this->attachment_id, self::KEY_ROOT_FILENAME, true );
    }

    /**
     * Set 'file_extension' meta value
     *
     * @param   string  $extension
     * @return  int|bool
     */
    public function set_file_extension( $extension ) {
        // Validate the value before saving
        if ( ! preg_match('/^png|jpe?g|gif$/i', $extension) ) {
            return false;
        }
        return update_post_meta( $this->attachment_id, self::KEY_FILE_EXT, $extension );
    }

    /**
     * Get 'file_extension' meta value
     *
     * @return  string
     */
    public function get_file_extension() {
        return get_post_meta( $this->attachment_id, self::KEY_FILE_EXT, true );
    }

    /**
     * Set 'aspectratio' meta value
     *
     * @param   mixed  $value
     * @return  int|bool
     */
    public function set_aspectratio( $value ) {
        if ( ! is_numeric($value) && ! is_array($value) ) {
            return false;
        }

        if ( is_array($value) ) {
            list( $width, $height ) = $value;
            $ratio = (float) $width / $height;
        } else {
            $ratio = (float) $value;
        }

        return update_post_meta( $this->attachment_id, self::KEY_ASPECT_RATIO, $ratio );
    }

    /**
     * Get 'aspectratio' meta value
     *
     * @return  string
     */
    public function get_aspectratio() {
        return get_post_meta( $this->attachment_id, self::KEY_ASPECT_RATIO, true );
    }

    /**
     * Set 'wpqd_enabled' meta value
     *
     * @param   bool  $value
     * @return  int|bool
     */
    public function set_enabled( $value ) {
        if ( $value === '' ) {
            return false;
        }
        return update_post_meta( $this->attachment_id, self::KEY_ENABLED, (bool) $value );
    }

    /**
     * Get 'wpqd_enabled' meta value
     *
     * @return  bool
     */
    public function get_enabled() {
        return (bool) get_post_meta( $this->attachment_id, self::KEY_ENABLED, true );
    }

    /**
     * Set 'wpqd_generated' meta value
     *
     * @param   bool  $value
     * @return  int|bool
     */
    public function set_imageset_generated( $value ) {
        if ( $value === '' ) {
            return false;
        }
        return update_post_meta( $this->attachment_id, self::KEY_IMAGESET_GENERATED, (bool) $value );
    }

    /**
     * Get 'wpqd_generated' meta value
     *
     * @return  bool
     */
    public function get_imageset_generated() {
        return (bool) get_post_meta( $this->attachment_id, self::KEY_IMAGESET_GENERATED, true );
    }

    /**
     * Set 'wpqd_max_breakpoint' meta value
     *
     * @param   int  $value
     * @return  int|bool
     */
    public function set_max_breakpoint( $value ) {
        if ( ! $value ) {
            return false;
        }
        return update_post_meta( $this->attachment_id, self::KEY_MAX_BREAKPOINT, (int) $value );
    }

    /**
     * Get 'wpqd_max_breakpoint' meta value
     *
     * @return  int
     */
    public function get_max_breakpoint() {
        return (int) get_post_meta( $this->attachment_id, self::KEY_MAX_BREAKPOINT, true );
    }

    /**
     * Set 'wpqd_under_min_limit' meta value
     *
     * @param   bool  $value
     * @return  int|bool
     */
    public function set_under_min_limit( $value ) {
        if ( $value === '' ) {
            return false;
        }
        return update_post_meta( $this->attachment_id, self::KEY_UNDER_MIN_LIMIT, (bool) $value );
    }

    /**
     * Get 'wpqd_under_min_limit' meta value
     *
     * @return  bool
     */
    public function get_under_min_limit() {
        return (bool) get_post_meta( $this->attachment_id, self::KEY_UNDER_MIN_LIMIT, true );
    }

    /**
     * Set 'wpqd_try_count' meta value
     *
     * @param   int  $value
     * @return  int|bool
     */
    public function set_try_count( $value = false ) {
        if ( ! $value ) {
            $value = 0;
        }
        return update_post_meta( $this->attachment_id, self::KEY_TRY_COUNT, (int) $value );
    }

    /**
     * Get 'wpqd_try_count' meta value
     *
     * @return  int
     */
    public function get_try_count() {
        return (int) get_post_meta( $this->attachment_id, self::KEY_TRY_COUNT, true );
    }

    /**
     * Set 'wpqd_try_start_time' meta value
     *
     * @param   int  $value
     * @return  int|bool
     */
    public function set_try_start_time( $value = false ) {
        if ( ! $value ) {
            $value = time();
        }
        return update_post_meta( $this->attachment_id, self::KEY_TRY_START_TIME, (int) $value );
    }

    /**
     * Get 'wpqd_try_start_time' meta value
     *
     * @return  int
     */
    public function get_try_start_time() {
        return (int) get_post_meta( $this->attachment_id, self::KEY_TRY_START_TIME, true );
    }

    /**
     * Set 'wpqd_try_start_time' meta value
     *
     * @param   int  $value
     * @return  int|bool
     */
    public function set_generated_time( $value = false ) {
        if ( ! $value ) {
            $value = time();
        }
        return update_post_meta( $this->attachment_id, self::KEY_GENERATED_TIME, (int) $value );
    }

    /**
     * Get 'wpqd_try_start_time' meta value
     *
     * @return  int
     */
    public function get_generated_time() {
        return (int) get_post_meta( $this->attachment_id, self::KEY_GENERATED_TIME, true );
    }


    /**
     * Get 'wpqd_ppng_leadcolor' meta value
     *
     * @return  int
     */
    public function get_ppng_leadcolor() {
        return get_post_meta( $this->attachment_id, self::KEY_PPNG_LEADCOLOR, true );
    }

    /**
     * Set 'wpqd_ppng_leadcolor' meta value
     *
     * @param   int  $value
     * @return  int|bool
     */
    public function set_ppng_leadcolor( $value = false ) {
        if ( ! $value ) {
            $value = time();
        }
        return update_post_meta( $this->attachment_id, self::KEY_PPNG_LEADCOLOR, $value );
    }

    /**
     * Set 'wpqd_img_cropped' meta value
     *
     * @param   bool  $value
     * @return  int|bool
     */
    public function set_cropped( $value ) {
        if ( $value === '' ) {
            return false;
        }
        return update_post_meta( $this->attachment_id, self::KEY_IMG_CROPPED, (bool) $value );
    }

    /**
     * Get 'wpqd_img_cropped' meta value
     *
     * @return  bool
     */
    public function get_cropped() {
        return (bool) get_post_meta( $this->attachment_id, self::KEY_IMG_CROPPED, true );
    }

    /**
     * Set 'wpqd_messages' meta value
     * 
     * @param   string  $value
     * @return  bool
     */
    public function set_message( $value = '' ) {
        return update_post_meta( $this->attachment_id, self::KEY_MESSAGE, (string) $value );
    }

    /**
     * Get 'wpqd_messages' meta value
     * 
     * @return  string
     */
    public function get_message() {
        return get_post_meta( $this->attachment_id, self::KEY_MESSAGE, true );
    }

    /**
     * Get imageset attribute data in JSON-encoded format
     *
     * @return  string
     */
    public function get_imageset_json() {
        return json_encode( $this->get_imageset_data() );
    }

    /**
     * Wrapper function for get_post_mime_type()
     * 
     * @return  bool|string
     */
    public function get_mime_type() {
        return get_post_mime_type( $this->attachment_id );
    }

    /**
     * Wrapper function for get_attached_file()
     * 
     * @return bool|string
     */
    public function get_filepath() {
        return get_attached_file( $this->attachment_id, false );
    }

    /**
     * Wrapper function for wp_get_attachment_metadata()
     * 
     * @return  bool|array
     */
    public function get_attachment_metadata() {
        return wp_get_attachment_metadata( $this->attachment_id, true );
    }

    /**
     * Get imageset attribute data
     *
     * @return  array
     */
    protected function get_imageset_data() {
        $data = array();
        $post_meta = get_post_meta( $this->attachment_id );

        foreach( array( self::KEY_ROOT_FILENAME, self::KEY_FILE_EXT ) as $attribute ) {
            if ( ! isset($post_meta[$attribute]) || empty( $post_meta[$attribute] ) ) {
                continue;
            }
            $data[$attribute] = array_shift( $post_meta[$attribute] );
        }
        return $data;
    }
}
