<?php
/**
 * WP QuickDraw is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * WP QuickDraw is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with WP QuickDraw. If not, see http://www.gnu.org/licenses/gpl.html.
 * 
 * @package     WP_QuickDraw
 * @author      hifipix
 * @copyright   2018 HifiPix, Inc.
 * @license     http://www.gnu.org/licenses/gpl.html  GNU GPL, Version 3
 * @version     1.0.1
 */
 
namespace Wpqd;

use \W3TC\Dispatcher;
use \Exception;
use \WP_Query;
use \Redux;
use \ReduxFrameworkInstances;

//-----------------------------------------
// Disallow Direct Plugin Browsing
//-----------------------------------------
defined( 'ABSPATH' ) or die( 'No script kiddies please!' );

class PluginState {

    /**
     * Path to directory where settings.js is saved
     *
     * @var string
     */
    protected $path;

    /**
     * Compatibility check errors list
     *
     * @var array
     */
    protected $errors = [];

    /**
     * @var string
     */
    const TRANSIENT_KEY_COMPAT = 'wpqd-plugin-admin-error';

    /**
     * @var string
     */
    const TRANSIENT_KEY_FILESYSTEM = 'wpqd-filesystem-error';

    /**
     * @var string
     */
    const OPTION_KEY_INIT = 'wpqd_image_import_initialized';

    /**
     * Register init_hooks
     *
     * @param   string  $file
     * @return  void
     */
    public function __construct( $file ) {
        $uploads = wp_upload_dir();
        $this->path = trailingslashit($uploads['basedir'] . WPQD_UPLOADS_FOLDER);

        register_activation_hook( $file, array( $this, 'activate' ) );
        register_deactivation_hook( $file, array( $this, 'deactivate' ) );

        add_action( 'admin_notices', array( $this, 'admin_error') );
        add_action( 'admin_notices', array( $this, 'admin_error_filesystem') );
        add_action( 'admin_notices', array( $this, 'wp_rocket_lazyload_check') );
        add_action( 'admin_notices', array( $this, 'w3tc_custom_file_cdn_check') );
        add_action( 'wp_ajax_wpqd_compatibility_check', array( $this, 'compatibility_check' ) );
        add_action( 'wp_ajax_wpqd_compatibility_check_ajax', array( $this, 'compatibility_check_ajax' ) );
        add_action( 'wp_ajax_wpqd_dismiss_admin_error', array( $this, 'dismiss_admin_error_ajax' ) );
        add_action( 'wp_ajax_wpqd_dismiss_admin_error_filesystem', array( $this, 'dismiss_admin_error_filesystem_ajax' ) );
        add_action( 'wp_ajax_wpqd_custom_file_cdn_update', array( $this, 'w3tc_custom_file_cdn_update' ) );
    }

    /**
     * On plugin activation / install
     *
     * @return  void
     */
    public function activate() {
        if ( $this->compatibility_check() ) {
            return true;
        }   
    }

    /**
     * On plugin deactivation
     * 
     * @return  void
     */
    public function deactivate() {
        $importer = BatchImporter::get_instance();
        $importer->pauseImport();
    }

    /**
     * Check for PHP module: GD Library
     *
     * @return  bool
     */
    public static function has_gd() {
        return (bool) (extension_loaded('gd') || function_exists('gd_info'));
    }

    /**
     * Check for PHP module: ImageMagick
     *
     * @return  bool
     */
    public static function has_imagick() {
        return (bool) (extension_loaded('imagick') || class_exists('Imagick'));
    }

    /**
     * Check for PHP module: DOM Library (provides XML support)
     *
     * @return  bool
     */
    public static function has_dom() {
        return (bool) (extension_loaded('dom') || class_exists('DOMDocument'));
    }

    /**
     * Run plugin compatibility check
     *
     * @return bool
     */
    public function compatibility_check() {
        WP_Filesystem(); // Init global $wp_filesystem
        global $wp_filesystem;

        $errors = [];

        if ( ! self::has_gd() && ! self::has_imagick() ) {
            $errors[] = 'Missing one of the following required PHP modules: GD Library (recommended) or ImageMagick';
        }

        if ( ! self::has_dom() ) {
            $errors[] = 'Missing required PHP module: DOM Library';
        }

        if ( ! $wp_filesystem->is_dir( $this->path ) ) {
            if ( ! $wp_filesystem->mkdir( $this->path ) ) {
                $errors[] = sprintf(
                    'Unable to create directory: %1$s. Please check directory permissions.',
                    $this->path
                );
            }
        }

        if ( $cron_error = $this->get_wp_cron_error() ) {
            $errors[] = $cron_error;
        }

        if ( ! empty($errors) ) {
            set_transient( self::TRANSIENT_KEY_COMPAT, $errors );
        } else {
            delete_transient( self::TRANSIENT_KEY_COMPAT );
        }

        $this->errors = $errors;
        $response = (bool) empty($errors);

        // After succesful compat check, initialize plugin and import
        if ($response) {
            $this->create_settings_json();
            $this->init_image_import();
        }

        return $response;
    }

    /**
     * Functionality for button in settings Tools > Compatibility Check
     *
     * @return  void
     */
    public function compatibility_check_ajax() {
        if ( $this->compatibility_check() ) {
            echo true;
        } else {
            echo json_encode( $this->errors );
        }
        wp_die();
    }

    /**
     * Delete transient when dismissing admin error (on "X" icon click)
     *
     * @return  void
     */
    public function dismiss_admin_error_ajax() {
        if ( delete_transient( self::TRANSIENT_KEY_COMPAT ) ) {
            echo true;
        } else {
            echo false;
        }
        wp_die();
    }

    /**
     * Delete transient when dismissing admin filesystem error (on "X" icon click)
     *
     * @return  void
     */
    public function dismiss_admin_error_filesystem_ajax() {
        if ( delete_transient( self::TRANSIENT_KEY_FILESYSTEM ) ) {
            echo true;
        } else {
            echo false;
        }
        wp_die();
    }

    /**
     * Print error message in admin
     *
     * @return void
     */
    public function admin_error() {
        $errors = get_transient( self::TRANSIENT_KEY_COMPAT );

        if ( ! $errors ) {
            return;
        }
        ?>
        <div class="notice notice-error is-dismissible" id="wpqd-warnings">
            <p><strong><?php _e( 'WP QuickDraw Warning(s):', 'wpqd' ); ?></strong></p>
            <ul>
            <?php foreach ( $errors as $error ): ?>
                <li><?php _e( $error, 'wpqd' ); ?></li>
            <?php endforeach; ?>
            </ul>
            <p>
                <a href="#" id="wpqd_compatibility_check" class="button button-primary button-large">
                    <?php _e( 'Recheck Compatibility', 'wpqd' ); ?>
                </a>
            </p>
        </div>
        <?php
    }

    /**
     * Print error message about filesystem settings in admin
     *
     * @return void
     */
    public function admin_error_filesystem() {
        $errors = get_transient( self::TRANSIENT_KEY_FILESYSTEM );

        if ( ! $errors ) {
            return;
        }
        ?>
        <div class="notice notice-error is-dismissible" id="wpqd-warnings-filesystem">
            <p><strong><?php _e( 'WP QuickDraw Warning(s):', 'wpqd' ); ?></strong></p>
            <ul>
            <?php foreach ( $errors as $error ): ?>
                <li><?php _e( $error, 'wpqd' ); ?></li>
            <?php endforeach; ?>
            </ul>
        </div>
        <?php
    }

    /**
     * Delete meta values in `wp_postmeta`
     *
     * @return  void
     */
    public function delete_metadata() {
        global $wpdb;

        $table = _get_meta_table( 'post' );
        $query = $wpdb->prepare(
            "DELETE FROM $table WHERE meta_key LIKE '%s'",
            'wpqd_%'
        );

        $wpdb->query( $query );
    }

    /**
     * Delete WPQD settings from database & static file
     *
     * @return  void
     */
    public function delete_settings() {
        delete_option( WPQD_OPTIONS );
        delete_option( self::OPTION_KEY_INIT );

        // For site options in Multisite
        if ( is_multisite() ) {
            delete_site_option( WPQD_OPTIONS );
            delete_site_option( self::OPTION_KEY_INIT );
        }

        // Delete settings.js file
        $this->delete_settings_json();
    }

    /**
     * Check for images "LazyLoad" in WP Rocket plugin (conflicts with WPQD)
     * 
     * @return  void
     */
    public function wp_rocket_lazyload_check() {
        if ( ! is_plugin_active( 'wp-rocket/wp-rocket.php' ) ) {
            return;
        }

        $settings = get_option( 'wp_rocket_settings' );
        if ( ! isset($settings['lazyload']) || $settings['lazyload'] != 1 ) {
            return;
        }
        ?>
        <div class="notice notice-warning is-dismissible">
            <p><strong><?php _e( 'WP QuickDraw Warning:', 'wpqd' ); ?></strong></p>
            <p>
                <?php printf(
                    __( 'You are currently using WP Rocket\'s image "LazyLoad" feature, which conflicts with WP QuickDraw. Please <a href="%s">disable this setting</a> to enjoy the functionality of WP QuickDraw.', 'wpqd' ),
                    get_site_url() . '/wp-admin/options-general.php?page=wprocket#media'
                ); ?>
            </p>
        </div>
        <?php
    }

    /**
     * Display admin warning about W3 Total Cache CDN "Custom File List"
     * 
     * @return  void
     */
    public function w3tc_custom_file_cdn_check() {
        if ( ! is_plugin_active( 'w3-total-cache/w3-total-cache.php' ) ) {
            return;
        }
    
        $config = Dispatcher::config();
        if ( ! $config->get_boolean( 'cdn.enabled' ) ) {
            return;
        }

        $custom_files = $config->get_array( 'cdn.custom.files' );
        $found = false;

        foreach ( $custom_files as $path ) {
            if ( preg_match('/^.*uploads.*\/wpqd.*$/i', $path) !== 1 ) {
                continue;
            }
            $found = true;
            break;
        }

        if ( $found ) {
            return;
        }

        ?>
        <div class="notice notice-warning is-dismissible">
            <p><strong><?php _e( 'WP QuickDraw Warning:', 'wpqd' ); ?></strong></p>
            <p>
                <?php printf(
                    __( 'You are currently utilizing a CDN through W3 Total Cache. <a href="%s">Please update your "Custom File List"</a> to include this folder in your CDN export: %s', 'wpqd' ),
                    get_site_url() . '/wp-admin/admin.php?page=w3tc_cdn#cdn_custom_files',
                    '{uploads_dir}' . WPQD_UPLOADS_FOLDER . '/*'
                ); ?>
                <p>
                <a href="#" id="wpqd_custom_file_cdn" class="button button-primary button-large">
                    <?php _e( 'Update settings automatically', 'wpqd' ); ?>
                </a>
                </p>
                <div id="wpqd_custom_file_cdn_feedback"></div>
            </p>
        </div>
        <?php
    }

    /**
     * Add WPQD uploads folder to W3 Total Cache CDN "Custom File List" setting
     * 
     * @return  void
     */
    public function w3tc_custom_file_cdn_update() {
        if ( ! is_plugin_active( 'w3-total-cache/w3-total-cache.php' ) ) {
            return;
        }
    
        $config = Dispatcher::config();
        $option_key = 'cdn.custom.files';

        $custom_files = $config->get_array( $option_key );
        $custom_files[] = '{uploads_dir}' . WPQD_UPLOADS_FOLDER . '/*';

        $config->set( $option_key, $custom_files );
        $config->save();

        echo json_encode( true );
        wp_die();
    }

    /**
     * Delete folder containing WPQD settings.js
     *
     * @return  void
     * @throws  Exception
     */
    protected function delete_settings_json() {
        WP_Filesystem(); // Init global $wp_filesystem
        global $wp_filesystem;

        if ( ! $wp_filesystem->delete( $this->path, true ) ) {
            throw new Exception( sprintf('Unable to delete: %1$s', $this->path) );
        }
    }

    /**
     * Create WPQD settings.js on plugin activation
     *
     * @return  void
     */
    protected function create_settings_json() {
        Redux::loadRedux( WPQD_OPTIONS );

        $redux = ReduxFrameworkInstances::get_instance( WPQD_OPTIONS );
        $redux->_register_settings(); // Saves default settings to database

        wpqd_save_config_file( get_option( WPQD_OPTIONS ) );
    }

    /**
     * Import all existing media library images on plugin activation
     * 
     * @return  void
     */
    protected function init_image_import() {
        global $wpdb;

        if ( $this->has_previous_import() ) {
            return;
        }

        $args = array(
            'post_type' => 'attachment',
            'post_status' => 'inherit',
            'post_mime_type' => 'image/jpeg,image/gif,image/jpg,image/png',
            'posts_per_page' => -1
        );

        $query_images = new WP_Query( $args );
        $image_rows = array();

        foreach ( $query_images->posts as $image ) {
            $image_rows[] = sprintf('(%1$s, "%2$s", %3$s)', 
                $image->ID,
                ImageMetadata::KEY_ENABLED,
                1
            );
        }

        // Delete existing values for WPQD enabled meta field
        $table = _get_meta_table( 'post' );
        $query_delete = $wpdb->prepare(
            "DELETE FROM $table WHERE meta_key = '%s'",
            ImageMetadata::KEY_ENABLED
        );
        $wpdb->query( $query_delete );

        // WPQD enable all images by default
        $query_insert = sprintf(
            'INSERT INTO %1$s (`post_id`, `meta_key`, `meta_value`) VALUES %2$s',
            $table,
            implode(', ', $image_rows)
        );
        $wpdb->query( $query_insert );

        // Initialize the image generation
        $batch_importer = BatchImporter::get_instance();
        $batch_importer->reimportAll();

        update_site_option( self::OPTION_KEY_INIT, true );
    }

    /**
     * Check for a previous or paused import on plugin activation
     * 
     * @return  bool
     */
    protected function has_previous_import() {
        $importer = BatchImporter::get_instance();
        if ( $importer->pauseImport() ) {
            return true;
        }

        return (bool) get_site_option( self::OPTION_KEY_INIT );
    }

    /**
     * Send test POST request to wp-cron.php to check if cron is running properly
     * 
     * @return  string|bool
     */
    protected function get_wp_cron_error() {
        global $wp_version;

        if ( defined( 'DISABLE_WP_CRON' ) && DISABLE_WP_CRON ) {
            return __( 'WP Cron is disabled. Please enable WP Cron so that image generation can proceed.', 'wpqd' );
        }
        
        $ssl_verify    = version_compare( $wp_version, 4.0, '<' );
        $doing_wp_cron = sprintf( '%.22F', microtime( true ) );
        $cron_request  = apply_filters( 'cron_request', array(
            'url'  => site_url( 'wp-cron.php?doing_wp_cron=' . $doing_wp_cron ),
            'key'  => $doing_wp_cron,
            'args' => array(
                'timeout'   => 10,
                'blocking'  => true,
                'sslverify' => apply_filters( 'https_local_ssl_verify', $ssl_verify ),
            ),
        ) );
        $cron_request['args']['timeout'] = 10;  // Force longer 10s timeout

        $result = wp_remote_post( $cron_request['url'], $cron_request['args'] );
        if ( is_wp_error( $result ) ) {
            return __( sprintf(
                'There was an error running WP Cron: "%s". WP Cron must be running properly for image generation to proceed.',
                $result->get_error_message()
            ), 'wpqd' );
        }
        
        $response_code = intval( wp_remote_retrieve_response_code( $result ) );
        if ( $response_code >= 300 ) {
            return __( sprintf( 
                'Unexpected HTTP response code "%s" from wp-cron.php. WP Cron must be running properly for image generation to proceed.', 
                $response_code 
            ), 'wpqd' );
        }
        
        return false;
    }
}
